import cupy


def polyvander(x, deg):
    """Computes the Vandermonde matrix of given degree.

    Args:
        x (cupy.ndarray): array of points
        deg (int): degree of the resulting matrix.

    Returns:
        cupy.ndarray: The Vandermonde matrix

    .. seealso:: :func:`numpy.polynomial.polynomial.polyvander`

    """
    deg = cupy.polynomial.polyutils._deprecate_as_int(deg, 'deg')
    if deg < 0:
        raise ValueError('degree must be non-negative')
    if x.ndim == 0:
        x = x.ravel()
    dtype = cupy.float64 if x.dtype.kind in 'biu' else x.dtype
    out = x ** cupy.arange(deg + 1, dtype=dtype).reshape((-1,) + (1,) * x.ndim)
    return cupy.moveaxis(out, 0, -1)


def polycompanion(c):
    """Computes the companion matrix of c.

    Args:
        c (cupy.ndarray): 1-D array of polynomial coefficients
            ordered from low to high degree.

    Returns:
        cupy.ndarray: Companion matrix of dimensions (deg, deg).

    .. seealso:: :func:`numpy.polynomial.polynomial.polycompanion`

    """
    [c] = cupy.polynomial.polyutils.as_series([c])
    deg = c.size - 1
    if deg == 0:
        raise ValueError('Series must have maximum degree of at least 1.')
    matrix = cupy.eye(deg, k=-1, dtype=c.dtype)
    matrix[:, -1] -= c[:-1] / c[-1]
    return matrix


def polyval(x, c, tensor=True):
    """
    Evaluate a polynomial at points x.

    If `c` is of length `n + 1`, this function returns the value

    .. math:: p(x) = c_0 + c_1 * x + ... + c_n * x^n

    The parameter `x` is converted to an array only if it is a tuple or a
    list, otherwise it is treated as a scalar. In either case, either `x`
    or its elements must support multiplication and addition both with
    themselves and with the elements of `c`.

    If `c` is a 1-D array, then `p(x)` will have the same shape as `x`.  If
    `c` is multidimensional, then the shape of the result depends on the
    value of `tensor`. If `tensor` is true the shape will be c.shape[1:] +
    x.shape. If `tensor` is false the shape will be c.shape[1:]. Note that
    scalars have shape (,).

    Trailing zeros in the coefficients will be used in the evaluation, so
    they should be avoided if efficiency is a concern.

    Parameters
    ----------
    x : array_like, compatible object
        If `x` is a list or tuple, it is converted to an ndarray, otherwise
        it is left unchanged and treated as a scalar. In either case, `x`
        or its elements must support addition and multiplication with
        with themselves and with the elements of `c`.
    c : array_like
        Array of coefficients ordered so that the coefficients for terms of
        degree n are contained in c[n]. If `c` is multidimensional the
        remaining indices enumerate multiple polynomials. In the two
        dimensional case the coefficients may be thought of as stored in
        the columns of `c`.
    tensor : boolean, optional
        If True, the shape of the coefficient array is extended with ones
        on the right, one for each dimension of `x`. Scalars have dimension 0
        for this action. The result is that every column of coefficients in
        `c` is evaluated for every element of `x`. If False, `x` is broadcast
        over the columns of `c` for the evaluation.  This keyword is useful
        when `c` is multidimensional. The default value is True.

    Returns
    -------
    values : ndarray, compatible object
        The shape of the returned array is described above.

    See Also
    --------
    numpy.polynomial.polynomial.polyval

    Notes
    -----
    The evaluation uses Horner's method.

    """
    c = cupy.array(c, ndmin=1, copy=False)
    if c.dtype.char in '?bBhHiIlLqQpP':
        # astype fails with NA
        c = c + 0.0
    if isinstance(x, (tuple, list)):
        x = cupy.asarray(x)
    if isinstance(x, cupy.ndarray) and tensor:
        c = c.reshape(c.shape + (1,)*x.ndim)

    c0 = c[-1] + x*0
    for i in range(2, len(c) + 1):
        c0 = c[-i] + c0*x
    return c0


def polyvalfromroots(x, r, tensor=True):
    """
    Evaluate a polynomial specified by its roots at points x.

    If `r` is of length `N`, this function returns the value

    .. math:: p(x) = \\prod_{n=1}^{N} (x - r_n)

    The parameter `x` is converted to an array only if it is a tuple or a
    list, otherwise it is treated as a scalar. In either case, either `x`
    or its elements must support multiplication and addition both with
    themselves and with the elements of `r`.

    If `r` is a 1-D array, then `p(x)` will have the same shape as `x`.  If `r`
    is multidimensional, then the shape of the result depends on the value of
    `tensor`. If `tensor` is ``True`` the shape will be r.shape[1:] + x.shape;
    that is, each polynomial is evaluated at every value of `x`. If `tensor` is
    ``False``, the shape will be r.shape[1:]; that is, each polynomial is
    evaluated only for the corresponding broadcast value of `x`. Note that
    scalars have shape (,).

    Parameters
    ----------
    x : array_like, compatible object
        If `x` is a list or tuple, it is converted to an ndarray, otherwise
        it is left unchanged and treated as a scalar. In either case, `x`
        or its elements must support addition and multiplication with
        with themselves and with the elements of `r`.
    r : array_like
        Array of roots. If `r` is multidimensional the first index is the
        root index, while the remaining indices enumerate multiple
        polynomials. For instance, in the two dimensional case the roots
        of each polynomial may be thought of as stored in the columns of `r`.
    tensor : boolean, optional
        If True, the shape of the roots array is extended with ones on the
        right, one for each dimension of `x`. Scalars have dimension 0 for this
        action. The result is that every column of coefficients in `r` is
        evaluated for every element of `x`. If False, `x` is broadcast over the
        columns of `r` for the evaluation.  This keyword is useful when `r` is
        multidimensional. The default value is True.

    Returns
    -------
    values : ndarray, compatible object
        The shape of the returned array is described above.

    See Also
    --------
    numpy.polynomial.polynomial.polyvalfroomroots
    """
    r = cupy.array(r, ndmin=1, copy=False)
    if r.dtype.char in '?bBhHiIlLqQpP':
        r = r.astype(cupy.double)
    if isinstance(x, (tuple, list)):
        x = cupy.asarray(x)
    if isinstance(x, cupy.ndarray):
        if tensor:
            r = r.reshape(r.shape + (1,)*x.ndim)
        elif x.ndim >= r.ndim:
            raise ValueError("x.ndim must be < r.ndim when tensor == False")
    return cupy.prod(x - r, axis=0)



import unittest

import numpy
import pytest

import cupy
from cupy import testing


def _generate_type_routines_input(xp, dtype, obj_type):
    dtype = numpy.dtype(dtype)
    if obj_type == 'dtype':
        return dtype
    if obj_type == 'specifier':
        return str(dtype)
    if obj_type == 'scalar':
        return dtype.type(3)
    if obj_type == 'array':
        return xp.zeros(3, dtype=dtype)
    if obj_type == 'primitive':
        return type(dtype.type(3).tolist())
    assert False


@testing.parameterize(
    *testing.product({
        'obj_type': ['dtype', 'specifier', 'scalar', 'array', 'primitive'],
    })
)
class TestCanCast(unittest.TestCase):

    @testing.for_all_dtypes_combination(names=('from_dtype', 'to_dtype'))
    @testing.numpy_cupy_equal()
    def test_can_cast(self, xp, from_dtype, to_dtype):
        from_obj = _generate_type_routines_input(xp, from_dtype, self.obj_type)
        ret = xp.can_cast(from_obj, to_dtype)
        assert isinstance(ret, bool)
        return ret


class TestCommonType(unittest.TestCase):

    @testing.numpy_cupy_equal()
    def test_common_type_empty(self, xp):
        ret = xp.common_type()
        assert type(ret) is type
        return ret

    @testing.for_all_dtypes(no_bool=True)
    @testing.numpy_cupy_equal()
    def test_common_type_single_argument(self, xp, dtype):
        array = _generate_type_routines_input(xp, dtype, 'array')
        ret = xp.common_type(array)
        assert type(ret) is type
        return ret

    @testing.for_all_dtypes_combination(
        names=('dtype1', 'dtype2'), no_bool=True)
    @testing.numpy_cupy_equal()
    def test_common_type_two_arguments(self, xp, dtype1, dtype2):
        array1 = _generate_type_routines_input(xp, dtype1, 'array')
        array2 = _generate_type_routines_input(xp, dtype2, 'array')
        ret = xp.common_type(array1, array2)
        assert type(ret) is type
        return ret

    @testing.for_all_dtypes()
    def test_common_type_bool(self, dtype):
        for xp in (numpy, cupy):
            array1 = _generate_type_routines_input(xp, dtype, 'array')
            array2 = _generate_type_routines_input(xp, 'bool_', 'array')
            with pytest.raises(TypeError):
                xp.common_type(array1, array2)


@testing.parameterize(
    *testing.product({
        'obj_type1': ['dtype', 'specifier', 'scalar', 'array', 'primitive'],
        'obj_type2': ['dtype', 'specifier', 'scalar', 'array', 'primitive'],
    })
)
class TestResultType(unittest.TestCase):

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @testing.numpy_cupy_equal()
    def test_result_type(self, xp, dtype1, dtype2):
        input1 = _generate_type_routines_input(xp, dtype1, self.obj_type1)
        input2 = _generate_type_routines_input(xp, dtype2, self.obj_type2)
        ret = xp.result_type(input1, input2)
        assert isinstance(ret, numpy.dtype)
        return ret



# flake8: NOQA
from cupyx.scipy.fft._fft import (
    fft, ifft, fft2, ifft2, fftn, ifftn,
    rfft, irfft, rfft2, irfft2, rfftn, irfftn,
    hfft, ihfft, hfft2, ihfft2, hfftn, ihfftn
)
from cupyx.scipy.fft._fft import (
    __ua_domain__, __ua_convert__, __ua_function__)
from cupyx.scipy.fft._fft import _scipy_150, _scipy_160
from cupyx.scipy.fft._fftlog import fht, ifht
from cupyx.scipy.fft._helper import next_fast_len  # NOQA
from cupy.fft import fftshift, ifftshift, fftfreq, rfftfreq
from cupyx.scipy.fftpack import get_fft_plan
from cupyx.scipy.fft._realtransforms import (
    dct, dctn, dst, dstn, idct, idctn, idst, idstn
)



import math

import cupyx.scipy.signal as signal  # noqa

import pytest
from cupy import testing


@testing.with_requires("scipy")
class TestC2D:

    @testing.numpy_cupy_allclose(scipy_name='scp', contiguous_check=False)
    def test_zoh(self, xp, scp):
        ac = xp.eye(2)
        bc = xp.full((2, 1), 0.5)
        cc = xp.array([[0.75, 1.0], [1.0, 1.0], [1.0, 0.25]])
        dc = xp.array([[0.0], [0.0], [-0.33]])
        dt_requested = 0.5

        c2d = scp.signal.cont2discrete
        ad, bd, cd, dd, dt = c2d((ac, bc, cc, dc), dt_requested, method='zoh')
        return ad, bd, cd, dd, dt

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_foh(self, xp, scp):
        ac = xp.eye(2)
        bc = xp.full((2, 1), 0.5)
        cc = xp.array([[0.75, 1.0], [1.0, 1.0], [1.0, 0.25]])
        dc = xp.array([[0.0], [0.0], [-0.33]])
        dt_requested = 0.5

        c2d = scp.signal.cont2discrete
        ad, bd, cd, dd, dt = c2d((ac, bc, cc, dc), dt_requested, method='foh')
        return ad, bd, cd, dd, dt

    @testing.numpy_cupy_allclose(scipy_name='scp', contiguous_check=False)
    def test_impulse(self, xp, scp):
        ac = xp.eye(2)
        bc = xp.full((2, 1), 0.5)
        cc = xp.array([[0.75, 1.0], [1.0, 1.0], [1.0, 0.25]])
        dc = xp.array([[0.0], [0.0], [0.0]])
        dt_requested = 0.5

        c2d = scp.signal.cont2discrete
        ad, bd, cd, dd, dt = c2d((ac, bc, cc, dc), dt_requested,
                                 method='impulse')
        return ad, bd, cd, dd, dt

    @testing.numpy_cupy_allclose(scipy_name='scp', contiguous_check=False)
    def test_gbt(self, xp, scp):
        ac = xp.eye(2)
        bc = xp.full((2, 1), 0.5)
        cc = xp.array([[0.75, 1.0], [1.0, 1.0], [1.0, 0.25]])
        dc = xp.array([[0.0], [0.0], [-0.33]])

        dt_requested = 0.5
        alpha = 1.0 / 3.0

        c2d = scp.signal.cont2discrete
        ad, bd, cd, dd, dt = c2d((ac, bc, cc, dc), dt_requested,
                                 method='gbt', alpha=alpha)
        return ad, bd, cd, dd, dt

    @testing.numpy_cupy_allclose(scipy_name='scp', contiguous_check=False)
    def test_euler(self, xp, scp):
        ac = xp.eye(2)
        bc = xp.full((2, 1), 0.5)
        cc = xp.array([[0.75, 1.0], [1.0, 1.0], [1.0, 0.25]])
        dc = xp.array([[0.0], [0.0], [-0.33]])
        dt_requested = 0.5

        c2d = scp.signal.cont2discrete
        ad, bd, cd, dd, dt = c2d((ac, bc, cc, dc), dt_requested,
                                 method='euler')
        return ad, bd, cd, dd, dt

    @testing.numpy_cupy_allclose(scipy_name='scp', contiguous_check=False)
    def test_backward_diff(self, xp, scp):
        ac = xp.eye(2)
        bc = xp.full((2, 1), 0.5)
        cc = xp.array([[0.75, 1.0], [1.0, 1.0], [1.0, 0.25]])
        dc = xp.array([[0.0], [0.0], [-0.33]])
        dt_requested = 0.5

        c2d = scp.signal.cont2discrete
        ad, bd, cd, dd, dt = c2d((ac, bc, cc, dc), dt_requested,
                                 method='backward_diff')
        return ad, bd, cd, dd, dt

    @pytest.mark.parametrize('dt_requested', [0.5, 1.0 / 3.0])
    @testing.numpy_cupy_allclose(scipy_name='scp', contiguous_check=False)
    def test_bilinear(self, xp, scp, dt_requested):
        ac = xp.eye(2)
        bc = xp.full((2, 1), 0.5)
        cc = xp.array([[0.75, 1.0], [1.0, 1.0], [1.0, 0.25]])
        dc = xp.array([[0.0], [0.0], [-0.33]])

        c2d = scp.signal.cont2discrete
        ad, bd, cd, dd, dt = c2d((ac, bc, cc, dc), dt_requested,
                                 method='bilinear')
        return ad, bd, cd, dd, dt

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_transferfunction(self, xp, scp):
        numc = xp.array([0.25, 0.25, 0.5])
        denc = xp.array([0.75, 0.75, 1.0])
        dt_requested = 0.5

        c2d = scp.signal.cont2discrete
        num, den, dt = c2d((numc, denc), dt_requested, method='zoh')
        return num, den, dt

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zerospolesgain(self, xp, scp):
        zeros_c = xp.array([0.5, -0.5])
        poles_c = xp.array([1.j / math.sqrt(2), -1.j / math.sqrt(2)])
        k_c = 1.0
        dt_requested = 0.5

        c2d = scp.signal.cont2discrete
        zeros, poles, k, dt = c2d((zeros_c, poles_c, k_c), dt_requested,
                                  method='zoh')
        return zeros, poles, k, dt

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_gbt_with_sio_tf_and_zpk(self, xp, scp):
        """Test method='gbt' with alpha=0.25 for tf and zpk cases."""
        # State space coefficients for the continuous SIO system.
        A = -1.0
        B = 1.0
        C = 1.0
        D = 0.5

        # The continuous transfer function coefficients.
        cnum, cden = scp.signal.ss2tf(A, B, C, D)

        # Continuous zpk representation
        cz, cp, ck = scp.signal.ss2zpk(A, B, C, D)

        h = 1.0
        alpha = 0.25

        # Compute the discrete tf using cont2discrete.
        c2d = scp.signal.cont2discrete
        c2dnum, c2dden, dt = c2d((cnum, cden), h, method='gbt', alpha=alpha)
        return c2dnum, c2dden, dt

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_discrete_approx(self, xp, scp):
        """
        Test that the solution to the discrete approximation of a continuous
        system actually approximates the solution to the continuous system.
        This is an indirect test of the correctness of the implementation
        of cont2discrete.
        """

        def u(t):
            return xp.sin(2.5 * t)

        a = xp.array([[-0.01]])
        b = xp.array([[1.0]])
        c = xp.array([[1.0]])
        d = xp.array([[0.2]])
        x0 = 1.0

        t = xp.linspace(0, 10.0, 101)
        dt = t[1] - t[0]
        u1 = u(t)

        # Use lsim to compute the solution to the continuous system.
        t, yout, xout = scp.signal.lsim((a, b, c, d), T=t, U=u1, X0=x0)

        # Convert the continuous system to a discrete approximation.
        dsys = scp.signal.cont2discrete((a, b, c, d), dt, method='bilinear')

        # Use dlsim with the pairwise averaged input to compute the output
        # of the discrete system.
        u2 = 0.5 * (u1[:-1] + u1[1:])
        t2 = t[:-1]
        td2, yd2, xd2 = scp.signal.dlsim(
            dsys, u=u2.reshape(-1, 1), t=t2, x0=x0)

        # ymid is the average of consecutive terms of the "exact" output
        # computed by lsim2.  This is what the discrete approximation
        # actually approximates.
        ymid = 0.5 * (yout[:-1] + yout[1:])

        return yd2, ymid

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simo_tf(self, xp, scp):
        # See gh-5753
        tf = ([[1, 0], [1, 1]], [1, 1])
        num, den, dt = scp.signal.cont2discrete(tf, 0.01)
        return num, den, dt

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_multioutput(self, xp, scp):
        ts = 0.01  # time step
        tf = ([[1, -3], [1, 5]], [1, 1])
        num, den, dt = scp.signal.cont2discrete(tf, ts)
        return num, den, dt

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_multioutput_1(self, xp, scp):
        ts = 0.01  # time step
        tf = ([[1, -3], [1, 5]], [1, 1])

        tf1 = (tf[0][0], tf[1])
        num1, den1, dt1 = scp.signal.cont2discrete(tf1, ts)
        return num1, den1, dt1

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_multioutput_2(self, xp, scp):
        ts = 0.01  # time step
        tf = ([[1, -3], [1, 5]], [1, 1])

        tf2 = (tf[0][1], tf[1])
        num2, den2, dt2 = scp.signal.cont2discrete(tf2, ts)
        return num2, den2, dt2


@testing.with_requires('scipy')
class TestC2dLti:

    @testing.numpy_cupy_allclose(scipy_name='scp', contiguous_check=False)
    def test_c2d_ss(self, xp, scp):
        # StateSpace
        A = xp.array([[-0.3, 0.1], [0.2, -0.7]])
        B = xp.array([[0], [1]])
        C = xp.array([[1, 0]])
        D = 0

        sys_ssc = scp.signal.lti(A, B, C, D)
        sys_ssd = sys_ssc.to_discrete(0.05)
        return sys_ssd.A, sys_ssd.B, sys_ssd.C, sys_ssd.D

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_c2d_tf(self, xp, scp):

        sys = scp.signal.lti([0.5, 0.3], [1.0, 0.4])
        sys = sys.to_discrete(0.005)
        return sys.num, sys.den


# Some test cases for checking the invariances.
# Array of triplets: (system, sample time, number of samples)
# here 'system' is a tuple to be wrapped into tfss(*system)
cases = [
    (([1, 1], [1, 1.5, 1]), 0.25, 10),
    (([1, 2], [1, 1.5, 3, 1]), 0.5, 10),
    ((0.1, [1, 1, 2, 1]), 0.5, 10),
]


@testing.with_requires('scipy')
class TestC2dInvariants:

    # Check that systems discretized with the impulse-invariant
    # method really hold the invariant
    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.parametrize("sys,sample_time,samples_number", cases)
    def test_impulse_invariant(self, xp, scp,
                               sys, sample_time, samples_number):
        time = xp.arange(samples_number) * sample_time
        sys = scp.signal.tf2ss(*sys)

        c2d = scp.signal.cont2discrete
        _, yout_cont = scp.signal.impulse(sys, T=time)
        _, yout_disc = scp.signal.dimpulse(
            c2d(sys, sample_time, method='impulse'), n=len(time))
        return yout_cont, yout_disc[0]

    # Step invariant should hold for ZOH discretized systems
    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.parametrize("sys,sample_time,samples_number", cases)
    def test_step_invariant(self, xp, scp, sys, sample_time, samples_number):
        time = xp.arange(samples_number) * sample_time
        sys = scp.signal.tf2ss(*sys)

        c2d = scp.signal.cont2discrete
        _, yout_cont = scp.signal.step(sys, T=time)
        _, yout_disc = scp.signal.dstep(
            c2d(sys, sample_time, method='zoh'), n=len(time))
        return yout_cont, yout_disc[0]

    # Linear invariant should hold for FOH discretized systems
    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.parametrize("sys,sample_time,samples_number", cases)
    def test_linear_invariant(self, xp, scp, sys, sample_time, samples_number):
        time = xp.arange(samples_number) * sample_time
        sys = scp.signal.tf2ss(*sys)

        c2d = scp.signal.cont2discrete
        _, yout_cont, _ = scp.signal.lsim(sys, T=time, U=time)
        _, yout_disc, _ = scp.signal.dlsim(
            c2d(sys, sample_time, method='foh'), u=time)
        return yout_cont, yout_disc
